import attr
import asyncio
import logging
from PySide6 import QtWidgets, QtGui, QtCore
import pyqtgraph as pg
import typing as tp

from .QWidgetWithCloseSignal import QMainWindowWithCloseSignal
from ..Signaler import Signal


from NaviNIBS.util import exceptionToStr
from NaviNIBS.util.Asyncio import asyncTryAndLogExceptionOnError

logger = logging.getLogger(__name__)


@attr.s(auto_attribs=True, eq=False, kw_only=True)
class RunnableAsApp:
    _appName: tp.Optional[str] = None
    _appAsyncPollPeriod = 0.01
    _doRunAsApp: bool = False
    _appLogEveryNLoops: tp.Optional[int] = None
    _theme: str = 'auto'  # auto, light, or dark

    _app: QtGui.QGuiApplication = attr.ib(init=False)
    _appIconPath: str | None = attr.ib(init=False, default=None)
    _Win: tp.Callable[..., QMainWindowWithCloseSignal] = attr.ib(default=QMainWindowWithCloseSignal)
    """
    Can specify Win to point to a different MainWindow class (e.g. for docking support)
    """
    _win: QMainWindowWithCloseSignal = attr.ib(init=False, default=None)
    _appIsClosing: bool = attr.ib(init=False, default=False)

    def __attrs_post_init__(self):
        if self._doRunAsApp:
            if self._appName is None:
                self._appName = self.__class__.__name__
            logger.info('Initializing %s' % (self._appName,))

            logger.debug('Initializing GUI window')
            self._app = pg.mkQApp(self._appName)

            if self._appIconPath is not None:

                import platform
                isWin = platform.system() == 'Windows'
                if isWin:
                    # workaround to show icon in Windows task bar
                    # based on https://stackoverflow.com/questions/1551605/how-to-set-applications-taskbar-icon-in-windows-7/1552105#1552105
                    import ctypes
                    myappid = f"NaviNIBS.{self._appName.replace(' ', '')}.subproduct.version"  # arbitrary string
                    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

                self._app.setWindowIcon(QtGui.QIcon(self._appIconPath))

            theme = self._theme
            if theme == 'auto':
                import darkdetect
                if darkdetect.isDark() and False:
                    theme = 'dark'
                else:
                    theme = 'light'
            match theme:
                case 'light':
                    pass  # do nothing
                case 'dark':
                    import qtawesome as qta
                    qta.dark(self._app)
                case _:
                    raise NotImplementedError

            self._win = self._Win()
            self._win.setWindowTitle(self._appName)
            self._win.sigAboutToClose.connect(self._onAppAboutToQuit)

    def _onAppAboutToQuit(self):
        logger.info('About to quit')
        self._appIsClosing = True

    @property
    def appIsClosing(self):
        return self._appIsClosing

    async def _runLoop(self):
        logger.debug('Running %s' % (self.__class__.__name__,))
        if self._appLogEveryNLoops is None:
            counter = 1
        else:
            counter = 0
        while not self._appIsClosing:
            if counter == 0:
                logger.debug('Main loop: process Qt events')
            try:
                self._app.processEvents()
            except Exception as e:
                logger.error(f'Exception while processing Qt events: {exceptionToStr(e)}')
                raise e
            if counter == 0:
                logger.debug('Main loop: async')
            await asyncio.sleep(self._appAsyncPollPeriod)
            if counter == 0:
                logger.debug('Main loop: end')
            # b/c doing QT event processing ourselves, for some reason the aboutToQuit signal is never emitted,
            # so check (infrequently) whether window has been closed and quit manually
            if self._appLogEveryNLoops is not None:
                counter = (counter + 1) % self._appLogEveryNLoops

    @classmethod
    def createAndRunAsTask(cls, *args, **kwargs):
        logger.debug('Creating %s' % (cls.__name__,))
        self = cls(*args, doRunAsApp=True, **kwargs)
        asyncio.create_task(asyncTryAndLogExceptionOnError(self._runLoop))
        return self

    @classmethod
    async def createAndRun_async(cls, *args, **kwargs):
        logger.debug('Creating %s' % (cls.__name__,))
        try:
            self = cls(*args, doRunAsApp=True, **kwargs)
        except Exception as e:
            raise e

        await self._runLoop()

    @classmethod
    def createAndRun(cls, *args, **kwargs):
        if True:
            from ..AsyncRunner import asyncioRunAndHandleExceptions
            asyncioRunAndHandleExceptions(cls.createAndRun_async, *args, **kwargs)
        else:
            asyncio.run(cls.createAndRun_async(*args, **kwargs), debug=True)


@attr.s(auto_attribs=True, eq=False)
class TestGUI(RunnableAsApp):
    _appName: str = 'Test GUI'

    def __attrs_post_init__(self):
        logger.info('Initializing TestGUI')

        super().__attrs_post_init__()

        self._wdgt = QtWidgets.QWidget()

        if self._doRunAsApp:
            self._win.setCentralWidget(self._wdgt)
            logger.info('Showing window')
            self._win.show()

    def _onAppAboutToQuit(self):
        self.cleanup()
        super()._onAppAboutToQuit()

    def cleanup(self):
        logger.info('Cleanup')


if __name__ == '__main__':
    TestGUI.createAndRun()


from __future__ import annotations

import asyncio

import attrs
from datetime import datetime
import logging
import nibabel as nib
from nibabel.affines import apply_affine
import numpy as np
import os
import pathlib
import pyvista as pv
import pyvistaqt as pvqt
import qtawesome as qta
from qtpy import QtWidgets, QtGui, QtCore
import shutil
import typing as tp

from . import MainViewPanel
from NaviNIBS.Navigator.GUI.Widgets.MRIViews import MRISliceView, MRI3DView
from NaviNIBS.Navigator.Model.Session import Session
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.GUI.QFileSelectWidget import QFileSelectWidget


logger = logging.getLogger(__name__)


@attrs.define()
class MRIPanel(MainViewPanel):
    _key: str = 'Set MRI'
    _icon: QtGui.QIcon = attrs.field(init=False, factory=lambda: qta.icon('mdi6.image'))
    _filepathWdgt: QFileSelectWidget = attrs.field(init=False)
    _views: tp.Dict[str, tp.Union[MRISliceView, MRI3DView]] = attrs.field(init=False, factory=dict)

    def __attrs_post_init__(self):
        super().__attrs_post_init__()

    def canBeEnabled(self) -> tuple[bool, str | None]:
        if self.session is None:
            return False, 'No session set'
        return True, None

    def _finishInitialization(self):
        # don't initialize computationally-demanding views until panel is activated (viewed)

        super()._finishInitialization()

        self._wdgt.setLayout(QtWidgets.QVBoxLayout())

        wdgt = QFileSelectWidget(browseMode='getOpenFilename',
                                 extFilters='Nifti (*.nii; *.nii.gz)')
        # TODO: set supported file formats to (.nii | .nii.gz) only
        wdgt.sigFilepathChanged.connect(self._onBrowsedNewFilepath)
        self._wdgt.layout().addWidget(wdgt)
        self._filepathWdgt = wdgt

        containerWdgt = QtWidgets.QWidget()
        containerWdgt.setSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        containerLayout = QtWidgets.QGridLayout()
        containerWdgt.setLayout(containerLayout)
        self._wdgt.layout().addWidget(containerWdgt)
        for iRow, iCol, key in ((0, 1, 'x'), (0, 0, 'y'), (1, 0, 'z'), (1, 1, '3D')):
            if key in ('x', 'y', 'z'):
                self._views[key] = MRISliceView(normal=key)
            elif key == '3D':
                self._views[key] = MRI3DView(label=key)
            else:
                raise NotImplementedError()

            self._views[key].sigSliceOriginChanged.connect(lambda key=key: self._onSliceOriginChanged(sourceKey=key))

            containerLayout.addWidget(self._views[key].wdgt, iRow, iCol)

        if self.session is not None:
            self._onPanelInitializedAndSessionSet()

    def _onSliceOriginChanged(self, sourceKey: str):
        for key, view in self._views.items():
            if key == sourceKey:
                continue
            view.sliceOrigin = self._views[sourceKey].sliceOrigin

    def _onSessionSet(self):
        super()._onSessionSet()

        if self._hasInitialized:
            self._onPanelInitializedAndSessionSet()

    def _onPanelInitializedAndSessionSet(self):
        self._updateFilepath()
        self._updateRelativeToPath()
        self.session.sigInfoChanged.connect(self._onSessionInfoChanged)
        self.session.MRI.sigFilepathChanged.connect(self._updateFilepath)

        for key, view in self._views.items():
            view.session = self.session

    def _updateFilepath(self):
        self._filepathWdgt.filepath = self.session.MRI.filepath

    def _onSessionInfoChanged(self, whatChanged: tp.Optional[list[str]] = None):
        if whatChanged is None or 'filepath' in whatChanged:
            self._updateRelativeToPath()

    def _updateRelativeToPath(self):
        self._filepathWdgt.showRelativeTo = os.path.dirname(self.session.filepath)
        self._filepathWdgt.showRelativePrefix = '<session>'

    def _onBrowsedNewFilepath(self, newFilepath: str):
        self.session.MRI.filepath = newFilepath



from __future__ import annotations

import shutil

import attrs
from datetime import datetime
import nibabel as nib
import json
import logging
import numpy as np
import os
import pandas as pd
import pyvista as pv
import tempfile
import typing as tp
from typing import ClassVar

from NaviNIBS.util.attrs import attrsAsDict
from NaviNIBS.util.Signaler import Signal
from NaviNIBS.util.numpy import array_equalish


logger = logging.getLogger(__name__)


@attrs.define()
class MRI:
    _filepath: tp.Optional[str] = None

    _data: tp.Optional[nib.Nifti1Image] = attrs.field(init=False, default=None)
    _dataAsUniformGrid: tp.Optional[pv.ImageData] = attrs.field(init=False, default=None)

    sigFilepathChanged: Signal = attrs.field(init=False, factory=Signal)
    sigDataChanged: Signal = attrs.field(init=False, factory=Signal)

    def __attrs_post_init__(self):
        self.sigFilepathChanged.connect(self._onFilepathChanged)
        self.validateFilepath(self._filepath)

    def loadCache(self):
        if not self.isSet:
            logger.warning('Load data requested, but no filepath set. Returning.')
            return

        logger.info('Loading image into cache from {}'.format(self.filepath))
        self._data = nib.load(self.filepath)

        if True:
            # create pyvista data object

            # get grid spacing from affine transform
            if np.all((self._data.affine[:-1, :-1] * (1-np.eye(3))) == 0):
                # for now, only on-diagonal transform supported
                gridSpacing = self._data.affine.diagonal()[:-1]
            else:
                raise NotImplementedError()

            if pv.__version__ <= '0.39.1':
                self._dataAsUniformGrid = pv.UniformGrid(
                    dims=self._data.shape)
            else:
                self._dataAsUniformGrid = pv.ImageData(
                    dimensions=self._data.shape,
                    spacing=gridSpacing,
                    origin=self._data.affine[:-1, 3]
                )

            self._dataAsUniformGrid.point_data['MRI'] = np.asanyarray(self.data.dataobj).ravel(order='F')

        self.sigDataChanged.emit()

    def clearCache(self):
        if self._data is None:
            return
        self._data = None
        self._dataAsUniformGrid = None
        self.sigDataChanged.emit()  # TODO: determine if necessary (since underlying uncached data didn't necessarily change)

    def _onFilepathChanged(self):
        with self.sigDataChanged.blocked():
            self.clearCache()
        self.sigDataChanged.emit()

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, newPath: str):
        if self._filepath == newPath:
            return
        self.validateFilepath(newPath)
        self._filepath = newPath
        self.sigFilepathChanged.emit()
        # TODO: here or with slots connected to sigDataChanged, make sure any cached MRI data or metadata is cleared/reloaded

    @property
    def isSet(self):
        return self._filepath is not None

    @property
    def data(self):
        if self.isSet and self._data is None:
            # data was not previously loaded, but it is available. Load now.
            self.loadCache()
        return self._data

    @property
    def dataAsUniformGrid(self):
        if self.isSet and self._data is None:
            # data was not previously loaded, but it is available. Load now.
            self.loadCache()
        return self._dataAsUniformGrid

    def asDict(self, filepathRelTo: str) -> tp.Dict[str, tp.Any]:
        d = attrsAsDict(self)
        if 'filepath' in d:
            d['filepath'] = os.path.relpath(d['filepath'], filepathRelTo)

        return d

    @classmethod
    def fromDict(cls, d: tp.Dict[str, tp.Any], filepathRelTo: str) -> MRI:
        # TODO: validate against schema
        if 'filepath' in d:
            d['filepath'] = os.path.join(filepathRelTo, d['filepath'])
            cls.validateFilepath(d['filepath'])
        return cls(**d)

    @classmethod
    def validateFilepath(cls, filepath: tp.Optional[str]) -> None:
        if filepath is None:
            return
        assert filepath.endswith('.nii') or filepath.endswith('.nii.gz')
        assert os.path.exists(filepath), 'File not found at {}'.format(filepath)

